from setuptools import setup, find_packages

setup(
    name="upowpy",
    version="1.0.10",
    author="upow",
    author_email="contact@upow.ai",
    description="upowpy is a Python library for interacting with uPow blockchain transactions, including sending funds, staking, voting, and managing nodes and validators on the blockchain",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/upowai/upowpy",
    packages=find_packages(),
    py_modules=["upowpy"],
    install_requires=[
        "asttokens==2.4.1",
        "base58==0.2.5",
        "certifi==2023.11.17",
        "charset-normalizer==2.0.12",
        "colorama==0.4.6",
        "executing==2.0.1",
        "fastecdsa==2.3.0",
        "icecream==2.1.3",
        "idna==3.6",
        "pygments==2.17.2",
        "requests==2.26.0",
        "six==1.16.0",
        "urllib3==1.26.18",
        "pickleDB~=0.9.2",
    ],
    python_requires=">=3.11",
)
