from fastecdsa import curve

ENDIAN = "little"
CURVE = curve.P256
SMALLEST = 100000000
MAX_SUPPLY = 18_884_643
VERSION = 1
MAX_BLOCK_SIZE_HEX = 4096 * 1024  # 4MB in HEX format, 2MB in raw bytes
MAX_INODES = 12
